@@ -17,27 +17,23 @@
 
 // Keep these in sync with other top-level files.
 #![warn(anonymous_parameters, bad_style, missing_docs)]
-#![warn(
-    unused,
-    unused_extern_crates,
-    unused_import_braces,
-    unused_qualifications
-)]
+#![warn(unused, unused_extern_crates, unused_import_braces, unused_qualifications)]
 #![warn(unsafe_code)]
 
-#[macro_use]
-extern crate failure;
-
-use failure::{Error, Fail, Fallible};
+use anyhow::{anyhow, Error, Result};
+use futures_lite::future::block_on;
 use getopts::Options;
+use std::cell::RefCell;
 use std::env;
 use std::fs::File;
+use std::io;
 use std::path::{Path, PathBuf};
 use std::process;
+use std::rc::Rc;
 
 /// Errors caused by the user when invoking this binary (invalid options or arguments).
-#[derive(Debug, Fail)]
-#[fail(display = "{}", message)]
+#[derive(Debug, thiserror::Error)]
+#[error("{message}")]
 struct UsageError {
     message: String,
 }
@@ -45,20 +41,14 @@ struct UsageError {
 impl UsageError {
     /// Creates a new usage error with `message`.
     fn new<T: Into<String>>(message: T) -> Self {
-        Self {
-            message: message.into(),
-        }
+        Self { message: message.into() }
     }
 }
 
 /// Flattens all causes of an error into a single string.
 fn flatten_causes(err: &Error) -> String {
-    err.iter_chain().fold(String::new(), |flattened, cause| {
-        let flattened = if flattened.is_empty() {
-            flattened
-        } else {
-            flattened + ": "
-        };
+    err.chain().fold(String::new(), |flattened, cause| {
+        let flattened = if flattened.is_empty() { flattened } else { flattened + ": " };
         flattened + &format!("{}", cause)
     })
 }
@@ -81,74 +71,77 @@ fn program_name(mut args: env::Args, default_name: &'static str) -> (String, env
 }
 
 /// Prints usage information for program `name` with `opts` following the GNU Standards format.
-fn help(name: &str, opts: &Options) -> Fallible<()> {
+fn help(name: &str, opts: &Options) -> Result<i32> {
     let brief = format!("Usage: {} [options] [program-file]", name);
     println!("{}", opts.usage(&brief));
     println!("Report bugs to: https://github.com/jmmv/endbasic/issues");
     println!("EndBASIC home page: https://github.com/jmmv/endbasic");
-    Ok(())
+    Ok(0)
 }
 
 /// Prints version information following the GNU Standards format.
-fn version() -> Fallible<()> {
+fn version() -> Result<i32> {
     println!("EndBASIC {}", env!("CARGO_PKG_VERSION"));
     println!("Copyright 2020 Julio Merino");
     println!("License Apache Version 2.0 <http://www.apache.org/licenses/LICENSE-2.0>");
-    Ok(())
+    Ok(0)
 }
 
 /// Computes the path to the directory where user programs live if `flag` is none; otherwise
 /// just returns `flag`.
-fn get_programs_dir(flag: Option<String>) -> Fallible<PathBuf> {
+fn get_programs_dir(flag: Option<String>) -> Result<PathBuf> {
     let dir = flag.map(PathBuf::from).or_else(|| {
-        dirs::document_dir()
-            .map(|d| d.join("endbasic"))
-            .or_else(|| {
-                // On Linux, dirs::document_dir() seems to return None whenever user-dirs.dirs is
-                // not present... which is suboptimal.  Compute a reasonable default based on the
-                // home directory.
-                dirs::home_dir().map(|h| h.join("Documents/endbasic"))
-            })
+        dirs::document_dir().map(|d| d.join("endbasic")).or_else(|| {
+            // On Linux, dirs::document_dir() seems to return None whenever user-dirs.dirs is
+            // not present... which is suboptimal.  Compute a reasonable default based on the
+            // home directory.
+            dirs::home_dir().map(|h| h.join("Documents/endbasic"))
+        })
     });
 
     // Instead of aborting on a missing programs directory, we could disable the LOAD/SAVE commands
     // when we cannot compute this folder, but that seems like hiding a corner case that is unlikely
     // to surface.  A good reason to do this, however, would be to allow the user to explicitly
     // disable this functionality to keep the interpreter from touching the disk.
-    ensure!(
-        dir.is_some(),
-        "Cannot compute default path to the Documents folder"
-    );
+    if dir.is_none() {
+        return Err(anyhow!("Cannot compute default path to the Documents folder"));
+    }
     Ok(dir.unwrap())
 }
 
-/// Implementation of the EndBASIC console to interact with stdin and stdout.
-#[derive(Default)]
-struct StdioConsole {}
-
 /// Executes the `path` program in a fresh machine.
-fn run<P: AsRef<Path>>(path: P) -> Fallible<()> {
-    let console = endbasic::repl::new_console();
-    let mut machine = endbasic::exec::MachineBuilder::default()
-        .add_builtins(endbasic::console::all_commands(console.clone()))
+fn run<P: AsRef<Path>>(path: P) -> endbasic_core::exec::Result<i32> {
+    let console = Rc::from(RefCell::from(endbasic::TextConsole::from_stdio()?));
+    let exit_code = Rc::from(RefCell::from(None));
+    let mut machine = endbasic_core::exec::MachineBuilder::default()
+        .add_builtin(Rc::from(endbasic_core::repl::ExitCommand::new(exit_code.clone())))
+        .add_builtins(endbasic_core::console::all_commands(console))
         .build();
 
     let mut input = File::open(path)?;
-    machine.exec(&mut input)
+    match block_on(machine.exec(&mut input)) {
+        Ok(()) => (),
+        Err(e) => {
+            if exit_code.borrow().is_some() {
+                if let endbasic_core::exec::Error::IoError(e) = e {
+                    debug_assert!(e.kind() == io::ErrorKind::UnexpectedEof);
+                }
+            } else {
+                return Err(e);
+            }
+        }
+    }
+    let exit_code = exit_code.borrow();
+    Ok(exit_code.unwrap_or(0))
 }
 
 /// Version of `main` that returns errors to the caller for reporting.
-fn safe_main(name: &str, args: env::Args) -> Fallible<()> {
+fn safe_main(name: &str, args: env::Args) -> Result<i32> {
     let args: Vec<String> = args.collect();
 
     let mut opts = Options::new();
     opts.optflag("h", "help", "show command-line usage information and exit");
-    opts.optopt(
-        "",
-        "programs-dir",
-        "directory where user programs are stored",
-        "PATH",
-    );
+    opts.optopt("", "programs-dir", "directory where user programs are stored", "PATH");
     opts.optflag("", "version", "show version information and exit");
     let matches = opts.parse(args)?;
 
@@ -163,27 +156,30 @@ fn safe_main(name: &str, args: env::Args) -> Fallible<()> {
     match matches.free.as_slice() {
         [] => {
             let programs_dir = get_programs_dir(matches.opt_str("programs-dir"))?;
-            Ok(endbasic::repl::run_repl_loop(&programs_dir)?)
+            Ok(endbasic::run_repl_loop(&programs_dir)?)
         }
-        [file] => run(file),
+        [file] => Ok(run(file)?),
         [_, ..] => Err(UsageError::new("Too many arguments").into()),
     }
 }
 
 fn main() {
     let (name, args) = program_name(env::args(), "endbasic");
-    if let Err(e) = safe_main(&name, args) {
-        if let Some(e) = e.downcast_ref::<UsageError>() {
-            eprintln!("Usage error: {}", e);
-            eprintln!("Type {} --help for more information", name);
-            process::exit(2);
-        } else if let Some(e) = e.downcast_ref::<getopts::Fail>() {
-            eprintln!("Usage error: {}", e);
-            eprintln!("Type {} --help for more information", name);
-            process::exit(2);
-        } else {
-            eprintln!("{}: {}", name, flatten_causes(&e));
-            process::exit(1);
+    match safe_main(&name, args) {
+        Ok(code) => process::exit(code),
+        Err(e) => {
+            if let Some(e) = e.downcast_ref::<UsageError>() {
+                eprintln!("Usage error: {}", e);
+                eprintln!("Type {} --help for more information", name);
+                process::exit(2);
+            } else if let Some(e) = e.downcast_ref::<getopts::Fail>() {
+                eprintln!("Usage error: {}", e);
+                eprintln!("Type {} --help for more information", name);
+                process::exit(2);
+            } else {
+                eprintln!("{}: {}", name, flatten_causes(&e));
+                process::exit(1);
+            }
         }
     }
 }
@@ -194,12 +190,12 @@ mod tests {
 
     #[test]
     fn flatten_causes_one() {
-        assert_eq!("the error", flatten_causes(&format_err!("the error")));
+        assert_eq!("the error", flatten_causes(&anyhow!("the error")));
     }
 
     #[test]
     fn flatten_causes_several() {
-        let err = Error::from(format_err!("first").context("second").context("and last"));
+        let err = anyhow!("first").context("second").context("and last");
         assert_eq!("and last: second: first", flatten_causes(&err));
     }
 }
